# This is done, for ease in imports
from .DutCacher import DutCacher
from .DutFetcher import DutFetcher
from .DutFilters import DutFilters
from .DutGrabber import DutGrabber
from .Exception import AutoFetchException



from typing import Callable
import pika
import json


class AsyncConsumer:

    def __init__(self, host: str, queue_name: str, callback: Callable[[dict], None]):
        self.host = host
        self.queue_name = queue_name
        self.callback = callback
        self.__connection: pika.adapters.blocking_connection.BlockingConnection = None
        self.__channel: pika.adapters.blocking_connection.BlockingChannel = None
        self.__connect()

    def __connect(self):
        try:
            self.__connection = pika.BlockingConnection(pika.ConnectionParameters(self.host))
            self.__channel = self.__connection.channel()
        except Exception as e:
            print(f"Error while connecting to RabbitMQ : {e}")
            return

    def __callback(self, channel, method, properties, body):
        try:
            print(f"Consuming message : {body.decode('utf-8')}")
            self.callback(json.loads(body.decode('utf-8')))
        except Exception as e:
            print(f"Error while consuming message : {body.decode('utf-8')} error : {e}")
        finally:
            channel.basic_ack(delivery_tag=method.delivery_tag)

    def start_consuming(self):
        try:
            self.__channel.basic_consume(queue=self.queue_name, on_message_callback=self.__callback, auto_ack=False)
            print('[*] Waiting for messages. To exit press CTRL+C')
            self.__channel.start_consuming()
        except KeyboardInterrupt as e:
            print("Exiting consumer...")
            self.__channel.stop_consuming()
            self.__connection.close()
        except Exception as e:
            print(f"Error occurred while consuming messages : {e}")
            self.__connection.close()




from packaging import version
import re


class DutFilters:
    __testing_env = False

    def __init__(self):
        pass

    @staticmethod
    def enable_test(enable: bool = True):
        DutFilters.__testing_env = enable

    @staticmethod
    def __print(message: str = ""):
        if DutFilters.__testing_env:
            print(message)

    @staticmethod
    def __normalize_version(v: str):
        """Converts version like 4.32.1F to 4.32.1.post1 for comparison."""
        if '-' in v:
            base_version = v.split('-', 1)[0]
            DutFilters.__print(f"[normalize_version] EOS Version contains '-' so, Base Version : {base_version}")
            normalized_base = DutFilters.__normalize_version(base_version)
            # For versions with '-', we only normalize the part before it.
            # We can return the normalized base along with the rest of the string if needed.
            # For now, let's just normalize the base.
            return normalized_base
        else:
            match = re.match(r'(\d+\.\d+\.\d+)([FM]?\+?)?', v)
            if not match:
                return v
            base, suffix = match.groups()
            DutFilters.__print(f"[normalize_version] Separating base and suffix, Base : {base} suffix : {suffix}")
            suffix = suffix or ''
            suffix_map = {'F': '.post1', 'M': '.post2', 'F+': '.post3', 'M+': '.post4'}
            return base + suffix_map.get(suffix, '')

    @staticmethod
    def is_valid_eos_version(allowed_ranges: dict[str, tuple[str, str]], v: str):
        """Check if EOS version is within the allowed range."""
        try:
            for major, (low, high) in allowed_ranges.items():
                if v.startswith(major):
                    v_norm = DutFilters.__normalize_version(v)
                    low_norm = DutFilters.__normalize_version(low)
                    high_norm = DutFilters.__normalize_version(high)
                    DutFilters.__print(f"[normalize_version] v_norm : {v_norm} low_norm : {low_norm} and high_norm : {high_norm}")
                    if version.parse(low_norm) <= version.parse(v_norm) <= version.parse(high_norm):
                        return True
        except Exception as e:
            print(f"Error while parsing version '{v}': {e}")
        return False

    @staticmethod
    def has_substr(pattern: str, main_str: str):
        return re.search(pattern, main_str)



import sqlite3


class DutCacher:

    def __init__(self, local_db: str, create_schema: bool = False, clean_up: bool = True):
        self.__connection: sqlite3.Connection = None
        self.__cursor: sqlite3.Cursor = None
        self.__schema = """
            CREATE TABLE device_cache (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                device TEXT NOT NULL,
                command TEXT NOT NULL,
                output TEXT NOT NULL,
                timestamp DATETIME DEFAULT CURRENT_TIMESTAMP
            );
        """
        self.local_db = local_db
        self.create_schema = create_schema
        self.__connect()
        if self.create_schema:
            self.__create_schema()
        if clean_up:
            self.__cleanup()

    def __connect(self):
        if self.__connection is None:
            try:
                self.__connection = sqlite3.connect(self.local_db)
                self.__cursor = self.__connection.cursor()
            except Exception as e:
                print(f"Exception while connecting to sqlite DB : {e}")

    def __create_schema(self):
        try:
            self.__cursor.execute(self.__schema)
            self.__connection.commit()
        except Exception as e:
            print(f"Exception while creating schema on sqlite : {e}")

    def __cleanup(self):
        try:
            self.__cursor.execute("DELETE FROM device_cache WHERE timestamp <= datetime('now', '-1 hour')")
            self.__connection.commit()
        except Exception as e:
            print(f"Error while cleaning up the database {e}")

    def get_cache(self, device: str, command: str):
        try:
            self.__cursor.execute(f"SELECT output FROM device_cache WHERE device = '{device}' AND command = '{command}' AND timestamp >= datetime('now', '-1 hour') ORDER BY timestamp")
            result = self.__cursor.fetchall()
            if len(result) == 0:
                return None
            else:
                return result[0][0]
        except Exception as e:
            print(f"Error while fetching data from sqlite {e}")

    def put_many_cache(self, multiple_records: []):
        try:
            self.__cursor.executemany("INSERT INTO device_cache (device, command, output) VALUES (?, ?, ?)", multiple_records)
            self.__connection.commit()
        except Exception as e:
            print(f"Error while inserting data into sqlite {e}")

    def put_single_query(self, query: str):
        try:
            self.__cursor.execute(query)
            self.__connection.commit()
        except Exception as e:
            print(f"Error while inserting data into sqlite {e}")

    def put_cache(self, device: str, command: str, output: str):
        try:
            self.__cursor.execute(f"INSERT INTO device_cache (device, command, output) values (?, ?, ?)", (device, command, output))
            self.__connection.commit()
        except Exception as e:
            print(f"Error while inserting data into sqlite {e}")

    def get_cursor(self):
        return self.__cursor




import pika
import json


class AsyncProducer:

    def __init__(self, host: str, queue_name: str, message: dict):
        self.host = host
        self.queue_name = queue_name
        self.message = message
        self.__connection: pika.adapters.blocking_connection.BlockingConnection = None
        self.__channel: pika.adapters.blocking_connection.BlockingChannel = None
        self.__connect()

    @staticmethod
    def create_and_publish(host: str, queue_name: str, message: dict):
        producer = AsyncProducer(host, queue_name, message)
        producer.publish_message_and_close(message)

    def __connect(self):
        try:
            self.__connection = pika.BlockingConnection(pika.ConnectionParameters(self.host))
            self.__channel = self.__connection.channel()
        except Exception as e:
            print(f"Error while connecting to RabbitMQ : {e}")
            return

    def __close_connection(self):
        self.__connection.close()

    def publish_message_and_close(self, message: dict, close: bool = True):
        try:
            self.__channel.queue_declare(queue=self.queue_name)
            message_body = json.dumps(message).encode('utf-8')
            self.__channel.basic_publish(exchange='',
                                         routing_key=self.queue_name,
                                         body=message_body,
                                         properties=pika.BasicProperties(
                                             content_type='application/json',
                                             delivery_mode=pika.DeliveryMode.Persistent
                                         ))
        except Exception as e:
            print(f"Error while publishing message to queue {self.queue_name} : {e}")
        finally:
            if close:
                self.__close_connection()
            return

    def publish_and_close(self):
        self.publish_message_and_close(self.message)

    def publish_message(self, message: dict):
        self.publish_message_and_close(message, False)

    def publish(self):
        self.publish_message_and_close(self.message, False)



import json
from typing import Callable
from concurrent.futures import ThreadPoolExecutor, as_completed
from collections import defaultdict
from AutoFetcher.DutCacher import DutCacher
from AutoFetcher.AsyncCacher.Consumer import AsyncConsumer
from AutoFetcher.AsyncCacher.Producer import AsyncProducer
from AutoFetcher.Exception import AutoFetchException

import pyeapi
import pyeapi.eapilib


class DutGrabber:
    def __init__(self, dut_list: list[str], single_task: Callable[[str, 'DutGrabber'], str],
                 local_db: str = None, create_schema: bool = False):
        """

        :param dut_list: The list of DUTs from DutFetcher or custom list.
        :param single_task: The user defined task in form of method to filter affected DUTs. A callable that processes
            a single device. The callable should accept the device name (str) and the DutGrabber instance, which
            provides methods to interact with the device.
        :param local_db: The sqlite database path as string from local machine (default it's None)
        :param create_schema: Boolean param that command the program whether to create schema automatically or not.
            (default it's False)
        """
        self.dutList = dut_list
        self.singleTaskHandler = single_task
        self.local_db = local_db
        self.create_schema = create_schema
        self.__host = None
        self.__queue_name = None
        self.__caching = False
        self.__consumer: AsyncConsumer = None

    def __patch_pyeapi_ciphers(self):
        try:
            import pyeapi.eapilib
        except ImportError:
            return
        connect_orig = pyeapi.eapilib.HttpsConnection.connect

        def connect(self):
            self._context.set_ciphers('DEFAULT@SECLEVEL=2')
            return connect_orig(self)

        pyeapi.eapilib.HttpsConnection.connect = connect

    def __fetch_dut(self, some_free_dut: str):
        result = self.singleTaskHandler(some_free_dut, self)
        return {
            "reason": result,
            "dut": some_free_dut
        }

    def enable_caching(self, host: str, queue_name: str, caching: bool = True):
        self.__caching = caching
        self.__host = host
        self.__queue_name = queue_name
        if self.__caching:
            # Just for cleaning the database for the first time
            # when user enable the caching
            DutCacher(self.local_db, clean_up=True)

    def execute_to_dut(self, device: str, commands=None):
        if commands is None:
            commands = ["show version detail"]
        try:
            results = []
            commands_not_cached = []
            final_results = []
            if not self.__caching:
                conn_to_dut = pyeapi.connect(host=f"{device}")
                results = conn_to_dut.execute(commands)
                final_results = results.get("result")
            else:
                # clean_up is set to false because we don't want to run delete query from each thread
                local_dut_cacher = DutCacher(self.local_db, clean_up=False)
                for command in commands:
                    cache = local_dut_cacher.get_cache(device, command)
                    if cache is not None:
                        results.append(json.loads(cache))
                    else:
                        commands_not_cached.append(command)
                        results.append("Fill")
                conn_to_dut = pyeapi.connect(host=f"{device}")
                not_cached_results = conn_to_dut.execute(commands_not_cached)
                last_fill = 0
                for result in results:
                    if result == "Fill":
                        if last_fill < len(not_cached_results):
                            AsyncProducer.create_and_publish(self.__host, self.__queue_name, {
                                "Device": device,
                                "Command": commands_not_cached[last_fill],
                                "Output": not_cached_results["result"][last_fill]
                            })
                            final_results.append(not_cached_results["result"][last_fill])
                            last_fill += 1
                    else:
                        final_results.append(result)
            return final_results
        except Exception as e:
            print(f"Error of type {type(e)} while executing commands {commands} to DUT {device}: {e}")
            raise AutoFetchException(f"[AUTO-FETCH] Exception of type {type(e)} : {e}")

    def fast_fetch_dut_with(self, all_free_dut: list[str], max_threads: int = 10):
        """
        This function can be used by users to execute multithreaded
        program that filters the DUT according to the callable
        **`self.singleTaskHandler`**

        Later, user can filter the results on their own.

        :param all_free_dut: The list of DUTs
        :param max_threads: The max threads that user might want. Default is `10`.
        :return: The mixed results. A list of results got from all threads.
        """
        result_collector = []
        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            tasks = {executor.submit(self.__fetch_dut, dut_name): dut_name for dut_name in all_free_dut}
            for task in as_completed(tasks):
                result = task.result()
                if result:
                    result_collector.append(result)
        return result_collector

    def fast_fetch_dut(self, max_threads: int = 10):
        """
        This function can be used by users to execute multithreaded
        program that filters the DUT according to the callable
        **`self.singleTaskHandler`**

        Later, user can filter the results on their own.

        :param max_threads: The max threads that user might want. Default is `10`.
        :return: The mixed results. A list of results got from all threads.
        """
        return self.fast_fetch_dut_with(self.dutList, max_threads)

    def fast_fetch_result_with(self, all_free_dut: list[str], max_threads: int = 10):
        """
        This function can be used by users to execute multithreaded
        program that filters the DUT according to the callable
        **`self.singleTaskHandler`** and you can pass the list of
        DUTs if you want.

        It is assumed that the callable **`self.singleTaskHandler`**
        returns the result of form string.

        :param all_free_dut: The list of DUTs
        :param max_threads: The max threads that user might want. Default is `10`.
        :return: The combined result.
        """
        mixed_results = self.fast_fetch_dut_with(all_free_dut, max_threads)
        return self.combine_results(mixed_results)

    def fast_fetch_result(self, max_threads: int = 10):
        """
        This function can be used by users to execute multithreaded
        program that filters the DUT according to the callable
        **`self.singleTaskHandler`.**

        It is assumed that the callable **`self.singleTaskHandler`**
        returns the result of form string.

        :param max_threads: The max threads that user might want. Default is `10`.
        :return: The combined result.
        """
        mixed_results = self.fast_fetch_dut_with(self.dutList, max_threads)
        return self.combine_results(mixed_results)

    @staticmethod
    def combine_results(mixed_results: list[{}]):
        """
        This function can be used by users to combine the mixed results
        of form `{'reason': <value>, 'dut': <value>}`
        that they might have received from the individual threads.

        :param mixed_results: The list of mixed results from each thread
        :return: The dictionary containing combined result
        """
        merged_result = defaultdict(list)
        for dut_info in mixed_results:
            merged_result[str(dut_info['reason'])].append(dut_info['dut'])
        return dict(merged_result)



from setuptools import setup, find_packages

with open('README.md', 'r') as file:
    description = file.read()

setup(
    name="AFAD",
    version="v0.1.0",
    author="Henil Rakeshbhai Mistry",
    author_email="henil@arista.com",
    description="Nothing but a human leveraging the power of threads to save time. See the below diagram and try to "
                "get it.",
    packages=find_packages(exclude=["AutoFetcher.Tests", "AutoFetcher.Tests.*", "Learning", "Resources", "ShowTechScripts", "ShowTechScripts.*", "Tests", "Tests.*"]),
    install_requires=[
        "pika>=1.3.2",
        "fabric>=3.2.2",
        "packaging>=25.0",
        "pyeapi>=1.0.4"
    ],
    long_description=description,
    long_description_content_type="text/markdown"
)



from fabric import Connection
import re


class DutFetcher:
    # This can be used to define Regular Expressions
    regularExpressions = {
        "DUT_RE": r"[a-zA-Z]+\d+"
    }

    def __init__(self, tac_server: str, tac_host: str, tac_password: str):
        """

        :param tac_server: The server which has Art Tool.
        :param tac_host: The arista username / hostname which will be authenticated to tac_server.
        :param tac_password: The password for tac_host for authentication.
        """
        self.__tacPassword = tac_password  # Stores the password for auth. of the tac_host
        self.__connection = None  # Store the connection to the tac_server
        self.tacServer = tac_server  # Stores the ip/name of the tac_server
        self.tacHost = tac_host  # Stores the name of the host/user for tac_server
        self.dutList = []  # Stores the list of DUTs processed from the output of last command!
        self.result = "NONE"  # Stores the unprocessed output of last command!
        self.allValidPools = []  # Stores the list of all valid pools
        # self.__find_all_valid_pools()

    def __connect_to_server(self):
        try:
            self.__connection = Connection(host=self.tacServer, user=self.tacHost,
                                           connect_kwargs={"password": f"{self.__tacPassword}"})
        except Exception as e:
            print(f"Error while connecting to Art server : {e}")

    def __fetch_command_output(self, command: str = "Art list --domain=all | greap free"):
        result = None
        try:
            if self.__connection is None:
                self.__connect_to_server()
            result = self.__connection.run(f"{command}", hide=True)
        except Exception as e:
            print(f"Error while collecting info from tac server: '{e}' while running command '{command}'")
        finally:
            if result is not None:
                return result.stdout
            else:
                return None

    def __find_all_valid_pools(self, command: str = "Art list --showpools"):
        self.result = self.__fetch_command_output(command)
        self.allValidPools = [pool_name for column_of_pools in self.result.splitlines()[2:] for pool_name in
                              column_of_pools.split()]

    def fetch_all_valid_pools(self, command: str = "Art list --showpools"):
        """
        This method can be used by users to run the command and fetch the
        list of all valid pools.

        :param command: The command that you want to run on Art server.
        :return: Returns the processed output (List of valid pools) from the command's output
        """
        self.__find_all_valid_pools(command)
        return self.allValidPools

    def fetch_duts(self, command: str = "Art list --domain=all | grep free"):
        """
        This method can be used by user to run the command and fetch the
        first column which usually holds the name of the DUT.

        :param command: The command that you want to run on Art server.
        :return: Returns the processed output (List of DUTs) from the command's output
        """
        self.result = self.__fetch_command_output(command)

        # This step is required because, command with "grep" won't have header in output...
        regex_command_containing_grep = r"grep"
        if re.search(regex_command_containing_grep, command):
            self.dutList = [
                line.split()[0]
                for line in self.result.strip().splitlines()
                if line.split()
            ]
        else:
            self.dutList = [line.split()[0] for line in self.result.strip().splitlines()[3:] if line.split()]

        # This step is required to filter extra rows for too long output from 'grab comment'
        self.dutList = [dut_name for dut_name in self.dutList if re.fullmatch(self.regularExpressions["DUT_RE"], dut_name)]
        return self.dutList

    def fetch_duts_from_pool(self, pool: str = "tac"):
        """
        This method can be used by user to run the command to fetch DUTs
        from pool as list of DUTs.

        :param pool: The name of the pool
        :return: Returns the dut column as list of DUTs from the pool
        """
        command = f"Art list --pool={pool}"
        return self.fetch_duts(command)

    def fetch_result(self, command: str = "Art list --domain=all | grep free"):
        """
        This method can be used by user to run the command and fetch the output directly!

        :param command: The command that you want to run on Art server.
        :return: Returns the actual/unprocessed output of the command
        """
        self.result = self.__fetch_command_output(command)
        return self.result

    def get_dut_list(self):
        return self.dutList

    def get_result(self):
        return self.result
